/*
 * Copyright 2012 Intel Corporation.
 *
 * This file is subject to the Intel Sample Source Code License. A copy
 * of the Intel Sample Source Code License is included.
 */

//Project Includes
#include "streaming_buffer_common.h"

//Standard Library Includes
#include <iostream>
#include <string>
#include <stdexcept>
#include <sstream>
#include <cstdlib>

#include <intel-coi/source/COIProcess_source.h>
#include <intel-coi/source/COIPipeline_source.h>
#include <intel-coi/source/COIEngine_source.h>
#include <intel-coi/source/COIBuffer_source.h>

//Namespaces
using namespace std;

// Simple class responsible for reading input from a file into buffers
class InputReader
{
public:
    static const size_t n_buffer = 8;
private:
    size_t _next_buffer;
    FILE  *_input_file;
    char   _last_char_of_previous_buffer;
    Buffer _buffer[n_buffer];
public:
    InputReader(const string &filename) : _next_buffer(0),
        _last_char_of_previous_buffer(' ')
    {
        // _last_char_of_previous_buffer is initialized to ' ' to handle
        // the case when the first character in a file needs to be capitalized
        _input_file = fopen(filename.c_str(),  "r");
        if(!_input_file)
            cerr << "Error opening input file \"" << filename << "\"" << endl;
    }

    ~InputReader()
    {
        fclose(_input_file);
    }

    // Reads from the files, returns true if eof has been reached
    bool read(Buffer *data)
    {
        Buffer &buffer = _buffer[_next_buffer];
        _next_buffer = (_next_buffer+1) % n_buffer;
		int n = (int)fread(buffer.begin(), 1, buffer.max_size(), _input_file);
        if(!n)
        {
            return true;
        }
        buffer.set_size(n);

        //Negative index indicates the last character of the previous buffer
        //-1 is the only valid negative index
        buffer[-1] = _last_char_of_previous_buffer;
        _last_char_of_previous_buffer = buffer[n-1];
        *data = buffer;
        return(n != buffer.max_size());//if not max size then eof was reached
    }
};

// Simple class responsible for writing buffer data into a file
class OutputWriter
{
private:
    FILE *_output_file;
public:
    OutputWriter(const string &filename)
    {
        _output_file = fopen(filename.c_str(),  "w");
        if(!_output_file)
            cerr << "Error opening output file \"" << filename << "\"" << endl;
    }

    ~OutputWriter()
    {
        fclose(_output_file);
    }

    // Writes buffer data to the file
    void write(Buffer *data)
    {
        fwrite(data->begin(), 1, data->size(), _output_file);
    }
};


// In this tutorial we will use COIProcess to create a software pipeline (seen
// below) using COIPipeline that will allow the host CPU, discrete compute
// device, and DMA engines to all stay busy at the same time.
// +---------+   +---------+   +---------+   +---------+
// | Source  |   | DMA to  |   | Sink    |   | DMA to  |
// | Compute |-->| Sink    |-->| Compute |-->| Source  |-->
// |         |   |         |   |         |   |         |
// +---------+   +---------+   +---------+   +---------+
//               +---------+   +---------+   +---------+   +---------+
//               | Source  |   | DMA to  |   | Sink    |   | DMA to  |
//               | Compute |-->| Sink    |-->| Compute |-->| Source  |-->
//               |         |   |         |   |         |   |         |
//               +---------+   +---------+   +---------+   +---------+
//                             +---------+   +---------+   +---------+   +---------+
//                             | Source  |   | DMA to  |   | Sink    |   | DMA to  |
//                             | Compute |-->| Sink    |-->| Compute |-->| Source  |-->
//                             |         |   |         |   |         |   |         |
//                             +---------+   +---------+   +---------+   +---------+
//                                           +---------+   +---------+   +---------+   +---------+
//                                           | Source  |   | DMA to  |   | Sink    |   | DMA to  |
//                                           | Compute |-->| Sink    |-->| Compute |-->| Source  |-->
//                                           |         |   |         |   |         |   |         |
//                                           +---------+   +---------+   +---------+   +---------+
// In this simple example host act as SOURCE and Intel(r) Xeon Phi(tm) card act as SINK. The
// "Source Compute" stage is responsible for loading text data from a file,
// which is transferred to the Intel(r) Xeon Phi(tm) card(acts as Sink) in the "DMA to Sink"
// stage. "Sink Compute" performs some basic transformation to the text, in
// this case capitalizing the first letter of a word, but potentially much more
// complex, i.e. encryption. The transformed buffer is sent back to
// the SOURCE in the "DMA to Source" stage and written to a file.
int main(int argc, char *argv[])
{
    string input_filename  = "input.txt";
    string output_filename = "output.txt";

    if(argc > 1) input_filename  = argv[1];
    if(argc > 2) output_filename = argv[2];

    // Get number of installed Intel(r) Xeon Phi(tm) devices
    uint32_t num_engines = 0;
    CHECK_RESULT(COIEngineGetCount(COI_ISA_MIC, &num_engines));

    if(num_engines == 0)
    {
        cerr << "Could not find any Intel(r) Xeon Phi(tm) devices. Is the environment "
            "configured properly?" << endl;
        cerr << "Press enter to exit" << endl;
        getchar();
        return(-1);
    }

    // Get a handle to the first Intel(r) Xeon Phi(tm) device so that we can use it later
    COIENGINE        engine;
    CHECK_RESULT(COIEngineGetHandle(COI_ISA_MIC, 0, &engine));

    // First, let's get the name of the binary we want to load

    const char* SINK_NAME = "streaming_buffer_sink_mic";

    // The following call creates a process on the sink.
    // We will automatically load any dependent libraries and run the
    // "main" function in the binary.
    COIPROCESS       proc;
    CHECK_RESULT( COIProcessCreateFromFile(
            engine,       // The engine to create the process on.
            SINK_NAME,    // The local path to the sink side binary to launch.
            0, NULL,      // argc and argv for the sink process.
            false, NULL,  // Environment variables to set for the sink process.
            true, NULL,   // Proxy
            1024*1024,    // The amount of memory to pre-allocate
                          // and register for use with COIBUFFERs.
            NULL,         // Path to search for dependencies
            &proc         // The resulting process handle.
        ));

    // Create a Pipeline to hold the Intel(r) Xeon Phi(tm) program that we want to run
    COIPIPELINE      pipe;
    CHECK_RESULT(COIPipelineCreate(proc, NULL, NULL, &pipe));
    printf("Created sink process %s and pipeline\n", SINK_NAME);

    // Obtain the handle to the function within the Intel(r) Xeon Phi(tm) program.
    COIFUNCTION      func;
    {
        const char*funcName[] = {"SINKCompute"};
        CHECK_RESULT(
        COIProcessGetFunctionHandles(proc, 1, funcName,
                            &func));
    }
    printf("Got handle to function SINKCompute\n");

    //Create a buffer written by Source and read by Sink
    COIBUFFER       toSinkBuffer;
    CHECK_RESULT(COIBufferCreate(sizeof(Buffer),COI_BUFFER_STREAMING_TO_SINK,
        NULL,NULL,1,&proc,&toSinkBuffer));

    //Create a buffer read by Source and written by Sink
    COIBUFFER       toSourceBuffer;
    CHECK_RESULT(COIBufferCreate(sizeof(Buffer),COI_BUFFER_STREAMING_TO_SOURCE,
                              NULL,NULL,1,&proc,&toSourceBuffer));
    printf("Create buffers\n");

    COIBUFFER   buffers[2];

    //Create input reader and output writer, give them the appropriate files to
    //work with
    InputReader input(input_filename);
    OutputWriter output(output_filename);

    bool eof = false;
    int  end = 4;

    // Prepare buffer list
    buffers[0]=toSinkBuffer;    // Input Data
    buffers[1]=toSourceBuffer;  // Output Data

    for(int i = 0; i < end; i++)
    {
        if(!eof)
        {
            COIMAPINSTANCE map_instance;
            Buffer *inputBufferData;
            // TO_SINK creates an a new version of the input buffer

            CHECK_RESULT(
                COIBufferMap(
                    toSinkBuffer,
                    0,                            // 0 indicates the entire
                                                  // buffer should be mapped.
                    0,                            // Map from the beginning
                                                  // of the buffer.
                    COI_MAP_WRITE_ENTIRE_BUFFER,  // Only need to read the
                                                  // buffer.
                    0, NULL,                      // No explicit dependencies.
                    NULL,                         // No completion event
                                                  // forces map to block.
                    &map_instance,                // the output handle used
                                                  // to unmap
                    (void**)&inputBufferData      // The pointer to the buffer
                                                  // data.
                ));

            {
                //Source reads data into an input buffer
                eof = input.read(inputBufferData);     // Once eof is true...
                end = i + 4;                           // we run the remaining
                                                       // stages of the pipeline
                //Print at every XTH iteration
                if(i%XTH == 0)
                {
                    printf("Mapped buffer and read input into it\n");
                }
            }

            // Source to Sink
            CHECK_RESULT(COIBufferUnmap(map_instance,0,NULL,NULL));
            if(i%XTH == 0)
            {
                printf("Unmapped buffer instance\n");
            }

            // Sink Compute
            // COIPipelineRunFunction inserts function in command queue,
            // function executed
            // asynchronously after input data is transferred
            COI_ACCESS_FLAGS flags[2] = {COI_SINK_READ, COI_SINK_WRITE_ENTIRE};
            const char *misc_data = "Misc Data";
            CHECK_RESULT(
                     COIPipelineRunFunction(
                     pipe,
                     func,
                     2, buffers, flags,
                     0, NULL,
                     misc_data, (uint16_t)strlen(misc_data)+1,
                     NULL,0,
                     NULL));

            if(i%XTH == 0)
            {
                printf("Called sink function SINKCompute(%s)\n", misc_data);
            }
        } // if (!eof)

        if(i >= 3) // Let the Sink get busy
        {
            COIMAPINSTANCE map_instance;
            Buffer *outputBufferData;
            //Nth Map call will the Map Buffer sent to Nth Runfunction.(N>0)
            CHECK_RESULT(
                COIBufferMap(
                    toSourceBuffer,
                    0,                            // 0 indicates the entire
                                                  // buffer should be mapped.
                    0,                            // Map from the beginning
                                                  // of the buffer.
                    COI_MAP_READ_ONLY,            // Only need to read the
                                                  // buffer.
                    0, NULL,                      // No explicit dependencies.
                    NULL,                         // No completion event
                                                  // forces map to block.
                    &map_instance,                // the output handle used
                                                  // to unmap
                    (void**)&outputBufferData));  // The pointer to the buffer
                                                  // data.
            {
                // Source  reads from the outputBuffer writes the data into
                // a Source file (Sink to Source stage)
                output.write(outputBufferData);
                if(i%XTH == 0)
                {
                    printf("Mapped output buffer and wrote data\n");
                }
            }

            // Sink to Source
            CHECK_RESULT( COIBufferUnmap(map_instance,0,NULL,NULL));
         }
    } // end for


    CHECK_RESULT(COIBufferDestroy(buffers[0]));
    CHECK_RESULT(COIBufferDestroy(buffers[1]));
    CHECK_RESULT(COIPipelineDestroy(pipe));

    printf("Destroyed buffers and pipelines\n");

    // Destroy the process
    CHECK_RESULT(
    COIProcessDestroy(
        proc,        // Process handle to be destroyed
        -1,          // Wait indefinitely until main() (on sink side) returns
        false,       // Don't force to exit. Let it finish executing
                     // functions enqueued and exit gracefully
        NULL,
        NULL ));

    printf("Exiting\n");

    return 0;
}
